# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

import os

import numpy as np

cimport cython
from libcpp.string cimport string

from .gravity cimport CGravityModel

from .error import GeographicErr
from ._utils import (
    as_contiguous_1d_llh,
    as_contiguous_1d_components,
    reshape_components,
)
from ._typing import PathType


cdef class GravityModel:
    r"""Model of the earth's gravity field.

    Evaluate the earth's gravity field according to a model.  The supported
    models treat only the gravitational field exterior to the mass of the
    earth.  When computing the field at points near (but above) the surface
    of the earth a small correction can be applied to account for the mass
    of the atmosphere above the point in question;
    see `The effect of the mass of the atmosphere
    <https://geographiclib.sourceforge.io/C++/doc/gravity.html#gravityatmos>`_.
    Determining the height of the geoid above the ellipsoid entails
    correcting for the mass of the earth above the geoid.
    The egm96 and egm2008 include separate correction terms to account for
    this mass.

    Definitions and terminology (from Heiskanen and Moritz, Sec 2-13):

    * :math:`V` = gravitational potential
    * :math:`\Phi` = rotational potential
    * :math:`W = V + \Phi = T + U` = total potential
    * :math:`V_{0}` = normal gravitation potential
    * :math:`U = V_{0} + \Phi` = total normal potential
    * :math:`T = W - U = V - V_{0}` = anomalous or disturbing potential
    * :math:`g = \nabla W = \gamma + \delta`
    * :math:`f = \nabla \Phi`
    * :math:`\Gamma = \nabla V_{0}`
    * :math:`\gamma = \nabla U`
    * :math:`\delta = \nabla T` = gravity disturbance vector =
      :math:`g_{P} - \gamma_{P}`
    * :math:`\delta g` = gravity disturbance = :math:`g_{P} - \gamma_{P}`
    * :math:`\Delta g` = gravity anomaly vector = :math:`g_{P} - \gamma_{Q}`
      here the line :math:`PQ` is perpendicular to ellipsoid and the potential
      at :math:`P` equals the normal potential at :math:`Q`
    * :math:`\Delta g` = gravity anomaly = :math:`g{P} - \gamma_{Q}`
    * :math:`(\xi, \eta)` deflection of the vertical, the difference in
      directions of :math:`g_{P}` and :math:`\gamma_{Q}, \xi = NS, \eta = EW`
    * :math:`X`, :math:`Y`, :math:`Z`, geocentric coordinates
    * :math:`x`, :math:`y`, :math:`z`, local cartesian coordinates used to
      denote the east, north and up directions.

    References:

    * W. A. Heiskanen and H. Moritz, Physical Geodesy (Freeman, San
      Francisco, 1967).

    See https://geographiclib.sourceforge.io/html/gravity.html.

    :param str name:
        the name of the model
    :param path:
        (optional) directory for data file
    :param max_degree:
        (optional) if non-negative, truncate the degree of the model this value
    :param max_order:
        (optional) if non-negative, truncate the order of the model this value
    :raises GeographicErr:
        if the data file cannot be found, is unreadable, or is corrupt
    :raises MemoryError:
        if the memory necessary for storing the model can't be
        allocated

    A filename is formed by appending ".egm" (World Gravity Model) to
    the name.  If `path` is specified (and is non-empty), then the file
    is loaded from directory, `path`.  Otherwise the path is given by
    :meth:`default_gravity_path`.

    This file contains the metadata which specifies the properties of
    the model.  The coefficients for the spherical harmonic sums are
    obtained from a file obtained by appending ".cof" to metadata file
    (so the filename ends in ".egm.cof").

    Example
    -------
    ::

        # Example of using the `geomodels.GravityModel` class.
        # This requires that the `egm96` gravity model be installed; see
        # https://geographiclib.sourceforge.io/C++/doc/gravity.html#gravityinst

        from geomodels import GravityModel

        grav = GravityModel("egm96")

        # Mt Everest
        lat = 27.99
        lon = 86.93
        h = 8820

        w, gx, gy, gz = grav.gravity(lat, lon, h)

        print(f"sum of the gravitational and centrifugal potentials (m**2/s**2): {w}")
        print(f"easterly component of the acceleration (m/s**2): {gx}")
        print(f"northerly component of the acceleration (m/s**2): {gy}")
        print(f"upward component of the acceleration (m/s**2): {gz} (usually negative")
    """
    cdef CGravityModel *_ptr

    def __cinit__(
        self,
        name: str | None = None,
        path: PathType = "",
        int max_degree=-1,
        int max_order=-1,
    ):
        cdef string c_name = CGravityModel.DefaultGravityName()
        cdef string c_path = os.fsencode(os.fspath(path))

        if name is not None:
            c_name = os.fsencode(name)

        try:
            with nogil:
                self._ptr = new CGravityModel(
                    c_name, c_path, max_degree, max_order
                )
        except RuntimeError as exc:
            raise GeographicErr(str(exc)) from exc

    def __dealloc__(self):
        del self._ptr

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _gravity(
        self, double[::1] vlat, double[::1] vlon, double[::1] vh
    ):
        cdef long size = vlat.size
        dtype = np.float64

        W = np.empty(shape=[size], dtype=dtype)
        gx = np.empty(shape=[size], dtype=dtype)
        gy = np.empty(shape=[size], dtype=dtype)
        gz = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vW = W
        cdef double[::1] vgx = gx
        cdef double[::1] vgy = gy
        cdef double[::1] vgz = gz

        cdef long i = 0
        with nogil:
            for i in range(size):
                vW[i] = self._ptr.Gravity(
                    vlat[i], vlon[i], vh[i], vgx[i], vgy[i], vgz[i]
                )

        return W, gx, gy, gz

    def gravity(self, lat, lon, h):
        """Evaluate the gravity at an arbitrary point above (or below)
        the ellipsoid.

        The function includes the effects of the earth's rotation.

        :param lat:
            the geographic latitude (degrees)
        :param lon:
            the geographic longitude (degrees)
        :param h:
            the height above the ellipsoid (meters)
        :returns:
            * `W`: the sum of the gravitational and centrifugal potentials
              ([m**2 / s**2])
            * `gx`: the easterly component of the acceleration ([m / s**2])
            * `gy`: the northerly component of the acceleration ([m / s**2])
            * `gz`: the upward component of the acceleration ([m / s**2]);
              this is usually negative
        """
        lat, lon, h, shape = as_contiguous_1d_llh(lat, lon, h, np.float64)
        W, gx, gy, gz = self._gravity(lat, lon, h)
        return reshape_components(shape, W, gx, gy, gz)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _disturbance(
        self, double[::1] vlat, double[::1] vlon, double[::1] vh
    ):
        cdef long size = vlat.size
        dtype = np.float64

        T = np.empty(shape=[size], dtype=dtype)
        deltax = np.empty(shape=[size], dtype=dtype)
        deltay = np.empty(shape=[size], dtype=dtype)
        deltaz = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vT = T
        cdef double[::1] vdeltax = deltax
        cdef double[::1] vdeltay = deltay
        cdef double[::1] vdeltaz = deltaz

        cdef long i = 0
        with nogil:
            for i in range(size):
                vT[i] = self._ptr.Disturbance(
                    vlat[i], vlon[i], vh[i], vdeltax[i], vdeltay[i], vdeltaz[i]
                )

        return T, deltax, deltay, deltaz

    def disturbance(self, lat, lon, h):
        """Evaluate the gravity disturbance vector at an arbitrary
        point above (or below) the ellipsoid.

        :param lat:
            the geographic latitude (degrees)
        :param lon:
            the geographic longitude (degrees)
        :param h:
            the height above the ellipsoid (meters)
        :returns:
            * `T`: the corresponding disturbing potential ([m**2 / s**2])
            * `delta_x`: the easterly component of the disturbance vector
              ([m / s**2])
            * `delta_y`: the northerly component of the disturbance vector
              ([m / s**2])
            * `delta_z`: the upward component of the disturbance vector
              ([m / s**2])
        """
        lat, lon, h, shape = as_contiguous_1d_llh(lat, lon, h, np.float64)
        T, deltax, deltay, deltaz = self._disturbance(lat, lon, h)
        return reshape_components(shape, T, deltax, deltay, deltaz)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _geoid_height(self, double[::1] vlat, double[::1] vlon):
        cdef long size = vlat.size
        dtype = np.float64

        h = np.empty(shape=[size], dtype=dtype)
        cdef double[::1] vh = h

        cdef long i = 0
        with nogil:
            for i in range(size):
                vh[i] = self._ptr.GeoidHeight(vlat[i], vlon[i])

        return h

    def geoid_height(self, lat, lon):
        """Evaluate the geoid height.

        :param lat:
            the geographic latitude (degrees)
        :param lon:
            the geographic longitude (degrees)
        :returns:
            `N` the height of the geoid above the reference ellipsoid
            (meters)

        Some approximations are made in computing the geoid height so that the
        results of the NGA codes are reproduced accurately.
        """
        lat, lon, shape = as_contiguous_1d_components(
            lat, lon, dtype=np.float64
        )
        h = self._geoid_height(lat, lon)
        return reshape_components(shape, h)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _spherical_anomaly(
            self, double[::1] vlat, double[::1] vlon, double[::1] vh):
        cdef long size = vlat.size
        dtype = np.float64

        Dg01 = np.empty(shape=[size], dtype=dtype)
        xi = np.empty(shape=[size], dtype=dtype)
        eta = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vDg01 = Dg01
        cdef double[::1] vxi = xi
        cdef double[::1] veta = eta

        cdef long i = 0
        with nogil:
            for i in range(size):
                self._ptr.SphericalAnomaly(
                    vlat[i], vlon[i], vh[i], vDg01[i], vxi[i], veta[i]
                )

        return Dg01, xi, eta

    def spherical_anomaly(self, lat, lon, h):
        """Evaluate the components of the gravity anomaly vector using
        the spherical approximation.

        :param lat:
            the geographic latitude (degrees)
        :param lon:
            the geographic longitude (degrees)
        :param h:
            the height above the ellipsoid (meters)
        :returns:
            * `Dg01`: the gravity anomaly ([m / s**2])
            * `xi`: the northerly component of the deflection of the
              vertical (degrees)
            * `eta`: the easterly component of the deflection of the
              vertical (degrees).

        The spherical approximation (see Heiskanen and Moritz,
        Sec 2-14) is used so that the results of the NGA codes are
        reproduced accurately.
        """
        lat, lon, h, shape = as_contiguous_1d_llh(lat, lon, h, np.float64)
        Dg01, xi, eta = self._spherical_anomaly(lat, lon, h)
        return reshape_components(shape, Dg01, xi, eta)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _w(self, double[::1] vx, double[::1] vy, double[::1] vz):
        cdef long size = vx.size
        dtype = np.float64

        out = np.empty(shape=[size], dtype=dtype)
        gx = np.empty(shape=[size], dtype=dtype)
        gy = np.empty(shape=[size], dtype=dtype)
        gz = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vout = out
        cdef double[::1] vgx = gx
        cdef double[::1] vgy = gy
        cdef double[::1] vgz = gz

        cdef long i = 0
        with nogil:
            for i in range(size):
                vout[i] = self._ptr.W(
                    vx[i], vy[i], vz[i], vgx[i], vgy[i], vgz[i]
                )

        return out, gx, gy, gz

    def w(self, x, y, z):
        """Evaluate the components of the acceleration due to gravity
        and the centrifugal acceleration in geocentric coordinates.

        :param x:
            geocentric coordinate of point (meters)
        :param y:
            geocentric coordinate of point (meters)
        :param z:
            geocentric coordinate of point (meters)
        :returns:
            * `W = V + Phi`: the sum of the gravitational and centrifugal
              potentials ([m**2 /  s**2])
            * `gx`: the x component of the acceleration ([m / s**2])
            * `gy`: the y component of the acceleration ([m / s**2])
            * `gz`: the z component of the acceleration ([m / s**2])
        """
        x, y, z, shape = as_contiguous_1d_components(
            x, y, z, labels=['x', 'y', 'z'], dtype=np.float64
        )
        out, gx, gy, gz = self._w(x, y, z)
        return reshape_components(shape, out, gx, gy, gz)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _v(self, double[::1] vx, double[::1] vy, double[::1] vz):
        cdef long size = vx.size
        dtype = np.float64

        out = np.empty(shape=[size], dtype=dtype)
        gx = np.empty(shape=[size], dtype=dtype)
        gy = np.empty(shape=[size], dtype=dtype)
        gz = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vout = out
        cdef double[::1] vgx = gx
        cdef double[::1] vgy = gy
        cdef double[::1] vgz = gz

        cdef long i = 0
        with nogil:
            for i in range(size):
                vout[i] = self._ptr.V(
                    vx[i], vy[i], vz[i], vgx[i], vgy[i], vgz[i]
                )

        return out, gx, gy, gz

    def v(self, x, y, z):
        """Evaluate the components of the gravity disturbance in
        geocentric coordinates.

        :param x:
            geocentric coordinate of point (meters)
        :param y:
            geocentric coordinate of point (meters)
        :param z:
            geocentric coordinate of point (meters)
        :returns:
            * `V = W - Phi`: the gravitational potential ([m**2 /  s**2])
            * `gx`: the x component of the acceleration ([m / s**2])
            * `gy`: the y component of the acceleration ([m / s**2])
            * `gz`: the z component of the acceleration ([m / s**2])
        """
        x, y, z, shape = as_contiguous_1d_components(
            x, y, z, labels=['x', 'y', 'z'], dtype=np.float64
        )
        out, gx, gy, gz = self._v(x, y, z)
        return reshape_components(shape, out, gx, gy, gz)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _t_components(
        self, double[::1] vx, double[::1] vy, double[::1] vz
    ):
        cdef long size = vx.size
        dtype = np.float64

        out = np.empty(shape=[size], dtype=dtype)
        deltax = np.empty(shape=[size], dtype=dtype)
        deltay = np.empty(shape=[size], dtype=dtype)
        deltaz = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vout = out
        cdef double[::1] vdeltax = deltax
        cdef double[::1] vdeltay = deltay
        cdef double[::1] vdeltaz = deltaz

        cdef long i = 0
        with nogil:
            for i in range(size):
                vout[i] = self._ptr.T(
                    vx[i], vy[i], vz[i], vdeltax[i], vdeltay[i], vdeltaz[i]
                )

        return out, deltax, deltay, deltaz

    def t_components(self, x, y, z):
        """Evaluate the components of the gravity disturbance in
        geocentric coordinates.

        :param x:
            geocentric coordinate of point (meters)
        :param y:
            geocentric coordinate of point (meters)
        :param z:
            geocentric coordinate of point (meters)
        :returns:
            * `T = W - U`: the disturbing potential (also called the
              anomalous potential) ([m**2 / s**2])
            * `delta_x`: the x component of the gravity disturbance
              ([m / s**2])
            * `delta_y`: the Y component of the gravity disturbance
              ([m / s**2])
            * `delta_z`: the Z component of the gravity disturbance
              ([m / s**2])
        """
        x, y, z, shape = as_contiguous_1d_components(
            x, y, z, labels=['x', 'y', 'z'], dtype=np.float64
        )
        out, gx, gy, gz = self._t_components(x, y, z)
        return reshape_components(shape, out, gx, gy, gz)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _t(self, double[::1] vx, double[::1] vy, double[::1] vz):
        cdef long size = vx.size
        dtype = np.float64

        out = np.empty(shape=[size], dtype=dtype)
        cdef double[::1] vout = out

        cdef long i = 0
        with nogil:
            for i in range(size):
                vout[i] = self._ptr.T(vx[i], vy[i], vz[i])

        return out

    def t(self, x, y, z):
        """Evaluate disturbing potential in geocentric coordinates.

        :param x:
            geocentric coordinate of point (meters)
        :param y:
            geocentric coordinate of point (meters)
        :param z:
            geocentric coordinate of point (meters)
        :returns:
            T = W - U, the disturbing potential (also called the
            anomalous potential) ([m**2 / s**2])
        """
        x, y, z, shape = as_contiguous_1d_components(
            x, y, z, labels=['x', 'y', 'z'], dtype=np.float64
        )
        out = self._t(x, y, z)
        return reshape_components(shape, out)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _u(self, double[::1] vx, double[::1] vy, double[::1] vz):
        cdef long size = vx.size
        dtype = np.float64

        out = np.empty(shape=[size], dtype=dtype)
        gammax = np.empty(shape=[size], dtype=dtype)
        gammay = np.empty(shape=[size], dtype=dtype)
        gammaz = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vout = out
        cdef double[::1] vgammax = gammax
        cdef double[::1] vgammay = gammay
        cdef double[::1] vgammaz = gammaz

        cdef long i = 0
        with nogil:
            for i in range(size):
                vout[i] = self._ptr.U(
                    vx[i], vy[i], vz[i], vgammax[i], vgammay[i], vgammaz[i]
                )

        return out, gammax, gammay, gammaz

    def u(self, x, y, z):
        """Evaluate the components of the acceleration due to normal
        gravity and the centrifugal acceleration in geocentric coordinates.

        :param x:
            geocentric coordinate of point (meters)
        :param y:
            geocentric coordinate of point (meters)
        :param z:
            geocentric coordinate of point (meters)
        :returns:
            * `U = V0 + Phi`:` the sum of the normal gravitational and
              centrifugal potentials ([m**2 / s**2])
            * `gamma_x`: the x component of the normal acceleration ([m / s**2])
            * `gamma_y`: the y component of the normal acceleration ([m / s**2])
            * `gamma_z`: the z component of the normal acceleration ([m / s**2])
        """
        x, y, z, shape = as_contiguous_1d_components(
            x, y, z, labels=['x', 'y', 'z'], dtype=np.float64
        )
        out, gammax, gammay, gammaz = self._u(x, y, z)
        return reshape_components(shape, out, gammax, gammay, gammaz)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _phi(self, double[::1] vx, double[::1] vy):
        cdef long size = vx.size
        dtype = np.float64

        out = np.empty(shape=[size], dtype=dtype)
        fx = np.empty(shape=[size], dtype=dtype)
        fy = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vout = out
        cdef double[::1] vfx = fx
        cdef double[::1] vfy = fy

        cdef long i = 0
        with nogil:
            for i in range(size):
                vout[i] = self._ptr.Phi(vx[i], vy[i], vfx[i], vfy[i])

        return out, fx, fy

    def phi(self, x, y):
        """Evaluate the centrifugal acceleration in geocentric coordinates.

        :param x:
            geocentric coordinate of point (meters)
        :param y:
            geocentric coordinate of point (meters)
        :returns:
            * `Phi`: the centrifugal potential ([m**2 / s**2])
            * `fx`: the x component of the centrifugal acceleration (m / s**2])
            * `fy`: the y component of the centrifugal acceleration (m / s**2])
        """
        x, y, shape = as_contiguous_1d_components(
            x, y, labels=['x', 'y'], dtype=np.float64
        )
        out, fx, fy = self._phi(x, y)
        return reshape_components(shape, out, fx, fy)

    # @TODO:
    # GravityCircle Circle(real lat, real h, unsigned caps=ALL) const except +
    # const NormalGravity& ReferenceEllipsoid() const

    def description(self) -> str:
        """The description of the gravity model.

        If the gravity model description is absent in the data file
        return "NONE".
        """
        return self._ptr.Description().decode('utf-8')

    def datetime(self) -> str:
        """Date of the gravity model.

        If absent, return "UNKNOWN".
        """
        return self._ptr.DateTime().decode('utf-8')

    def gravity_file(self) -> str:
        """Full file name used to load the gravity model."""
        return self._ptr.GravityFile().decode('utf-8')

    def gravity_model_name(self) -> str:
        """Return the "name" used to load the gravity model.

        "name" is the first argument of the constructor, but this may
        be overridden by the model file).
        """
        return self._ptr.GravityModelName().decode('utf-8')

    def gravity_model_directory(self) -> str:
        """Return the directory used to load the gravity model."""
        return self._ptr.GravityModelDirectory().decode('utf-8')

    def equatorial_radius(self) -> float:
        """The equatorial radius of the ellipsoid (meters)."""
        return self._ptr.EquatorialRadius()

    def flattening(self) -> float:
        """The flattening of the ellipsoid."""
        return self._ptr.Flattening()

    def mass_constant(self) -> float:
        """The mass constant of the model ([GM] = m**3 /s**2).

        It is the product of `G` the gravitational constant and `M` the
        mass of the earth (usually including the mass of the earth's
        atmosphere).
        """
        return self._ptr.MassConstant()

    def reference_mass_constant(self) -> float:
        """The mass constant of the reference ellipsoid.

        [GM] = m**3 / s**2.
        """
        return self._ptr.ReferenceMassConstant()

    def angular_velocity(self) -> float:
        r"""The angular velocity of the model and the reference ellipsoid.

        :math:`[\omega] = rad/s`
        """
        return self._ptr.AngularVelocity()

    def degree(self) -> int:
        """The maximum degree of the components of the model (`Nmax`)."""
        return self._ptr.Degree()

    def order(self) -> int:
        """The maximum order of the components of the model (`Mmax`)."""
        return self._ptr.Order()

    @staticmethod
    def default_gravity_path() -> str:
        """the default path for gravity model data files.

        This is the value of the environment variable
        `GEOGRAPHICLIB_GRAVITY_PATH`, if set; otherwise, it is
        `$GEOGRAPHICLIB_DATA/gravity` if the environment variable
        `GEOGRAPHICLIB_DATA` is set; otherwise, it is a compile-time
        default.
        """
        return CGravityModel.DefaultGravityPath().decode('utf-8')

    @staticmethod
    def default_gravity_name() -> str:
        """The default name for the gravity model.

        This is the value of the environment variable
        `GEOGRAPHICLIB_GRAVITY_NAME`, if set; otherwise, it is "egm96".
        The :class:`GravityModel` class does not use this function;
        it is just provided as a convenience for a calling program when
        constructing a :class:`GravityModel` object.
        """
        return CGravityModel.DefaultGravityName().decode('utf-8')



# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

from libcpp.string cimport string

from .common cimport real


cdef extern from "GeographicLib/GravityModel.hpp" namespace "GeographicLib" nogil:
    cdef cppclass CGravityModel "GeographicLib::GravityModel":
        # ctypedef enum mask:
        #     NONE,
        #     GRAVITY,
        #     DISTURBANCE,
        #     DISTURBING_POTENTIAL,
        #     SPHERICAL_ANOMALY,
        #     GEOID_HEIGHT,
        #     ALL

        CGravityModel(
            const string& name, const string& path, int Nmax, int Mmax
        ) except +

        real Gravity(
            real lat, real lon, real h, real& gx, real& gy, real& gz
        ) const
        real Disturbance(
            real lat,
            real lon,
            real h,
            real& deltax,
            real& deltay,
            real& deltaz,
        ) const
        real GeoidHeight(real lat, real lon) const
        void SphericalAnomaly(
            real lat, real lon, real h, real& Dg01, real& xi, real& eta
        ) const
        real W(real X, real Y, real Z, real& gX, real& gY, real& gZ) const
        real V(real X, real Y, real Z, real& GX, real& GY, real& GZ) const
        real T(
            real X, real Y, real Z, real& deltaX, real& deltaY, real& deltaZ
        ) const
        real T(real X, real Y, real Z) const
        real U(
            real X, real Y, real Z, real& gammaX, real& gammaY, real& gammaZ
        ) const
        real Phi(real X, real Y, real& fX, real& fY) const

        # GravityCircle Circle(
        #     real lat, real h, unsigned caps=ALL
        # ) const except +
        # const NormalGravity& ReferenceEllipsoid() const

        const string& Description() const
        const string& DateTime() const
        const string& GravityFile() const
        const string& GravityModelName() const
        const string& GravityModelDirectory() const

        real EquatorialRadius() const
        real MassConstant() const
        real ReferenceMassConstant() const
        real AngularVelocity() const
        real Flattening() const

        int Degree() const
        int Order() const

        @staticmethod
        string DefaultGravityPath()

        @staticmethod
        string DefaultGravityName()



import os
import shutil
import pathlib
import tempfile
import unittest
from unittest import mock
from urllib.parse import urlsplit

import geomodels.data


class DataPathTestCase(unittest.TestCase):
    def test_default_data_path(self):
        path = geomodels.data.get_default_data_path()
        self.assertTrue(os.path.exists(path))

    def test_default_data_path_from_env(self):
        old_env = os.environ.get("GEOGRAPHICLIB_DATA")
        dummy_path = "/dummy/path"
        os.environ["GEOGRAPHICLIB_DATA"] = dummy_path

        try:
            path = geomodels.data.get_default_data_path()
            self.assertEqual(path, dummy_path)
        finally:
            if old_env is None:
                del os.environ["GEOGRAPHICLIB_DATA"]
            else:
                os.environ["GEOGRAPHICLIB_DATA"] = old_env


class UrlTestCase(unittest.TestCase):
    MODEL_TYPES = [
        geomodels.data.EGeoidModel,
        geomodels.data.EGravityModel,
        geomodels.data.EMagneticModel,
    ]

    def test_get_base_url(self):
        self.assertIsInstance(geomodels.data.get_base_url(), str)

    def test_get_model_url(self):
        base_url = geomodels.data.get_base_url()
        archive_extension = geomodels.data.EArchiveType.BZ2.value
        model: geomodels.data.GenericModelType
        for model_type in self.MODEL_TYPES:
            for model in model_type:  # type: ignore[assignment]
                with self.subTest(model=model):
                    url = geomodels.data.get_model_url(model)
                    self.assertIsInstance(url, str)
                    self.assertTrue(url.startswith(base_url))
                    urlobj = urlsplit(url)
                    self.assertIn(model.get_model_type().value, urlobj.path)
                    self.assertIn(model.value, urlobj.path)
                    self.assertTrue(urlobj.path.endswith(archive_extension))

    def test_get_model_url_with_base_url(self):
        base_url = "https://dummy.url"
        archive_extension = geomodels.data.EArchiveType.BZ2.value
        model: geomodels.data.GenericModelType
        for model_type in self.MODEL_TYPES:
            for model in model_type:  # type: ignore[assignment]
                with self.subTest(model=model):
                    url = geomodels.data.get_model_url(
                        model, base_url=base_url
                    )
                    self.assertIsInstance(url, str)
                    self.assertTrue(url.startswith(base_url))
                    urlobj = urlsplit(url)
                    self.assertIn(model.get_model_type().value, urlobj.path)
                    self.assertIn(model.value, urlobj.path)
                    self.assertTrue(urlobj.path.endswith(archive_extension))

    def test_get_model_url_with_archive_type(self):
        base_url = geomodels.data.get_base_url()
        archive_type = geomodels.data.EArchiveType.ZIP
        archive_extension = archive_type.value
        model: geomodels.data.GenericModelType
        for model_type in self.MODEL_TYPES:
            for model in model_type:  # type: ignore[assignment]
                with self.subTest(model=model):
                    url = geomodels.data.get_model_url(
                        model, archive_type=archive_type
                    )
                    self.assertIsInstance(url, str)
                    self.assertTrue(url.startswith(base_url))
                    urlobj = urlsplit(url)
                    self.assertIn(model.get_model_type().value, urlobj.path)
                    self.assertIn(model.value, urlobj.path)
                    self.assertTrue(urlobj.path.endswith(archive_extension))


class DownloadTestCase(unittest.TestCase):
    DATA = "dummydata"

    def setUp(self) -> None:
        self.datadir = pathlib.Path(tempfile.mkdtemp())
        self.srcfile = self.datadir / "filename.txt"
        with open(self.srcfile, "w") as fd:
            fd.write(self.DATA)

    def tearDown(self) -> None:
        shutil.rmtree(self.datadir)

    def test_download_file_uri(self):
        dstdir = self.datadir / "out"
        dstdir.mkdir()

        dstfile = dstdir / self.srcfile.name
        self.assertFalse(dstfile.exists())

        url = self.srcfile.as_uri()
        geomodels.data.download(url, dstdir, progress=False)

        self.assertTrue(dstfile.exists())
        with open(dstfile) as fd:
            data = fd.read()

        self.assertEqual(data, self.DATA)

    @unittest.skipIf(
        not isinstance(pathlib.Path(), pathlib.PosixPath),
        reason="not on a posix platform.",
    )
    def test_download_file(self):
        dstdir = self.datadir / "out"
        dstdir.mkdir()

        dstfile = dstdir / self.srcfile.name
        self.assertFalse(dstfile.exists())

        url = str(self.srcfile)
        geomodels.data.download(url, dstdir, progress=False)

        self.assertTrue(dstfile.exists())
        with open(dstfile) as fd:
            data = fd.read()

        self.assertEqual(data, self.DATA)


class InstallTestCase(unittest.TestCase):
    MODEL_TYPES = [
        geomodels.data.EGeoidModel,
        geomodels.data.EGravityModel,
        geomodels.data.EMagneticModel,
    ]

    def setUp(self) -> None:
        self.download_patcher = mock.patch(
            "geomodels.data.download", return_value="dummy"
        )
        self.unpack_archive_patcher = mock.patch("shutil.unpack_archive")
        self.download_mock = self.download_patcher.start()
        self.unpack_archive_mock = self.unpack_archive_patcher.start()

    def tearDown(self) -> None:
        self.download_patcher.stop()
        self.unpack_archive_patcher.stop()

    def test_install_one_model(self):
        model = geomodels.data.EGravityModel.EGM96
        url = geomodels.data.get_model_url(model)

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        self.download_mock.assert_called_once_with(
            url, mock.ANY, progress=False
        )
        self.unpack_archive_mock.assert_called_once_with(
            mock.ANY, extract_dir=datadir
        )

    def test_install_one_model_type(self):
        model = geomodels.data.EModelType.GRAVITY
        n_models = len(geomodels.data.EGravityModel)

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_all_models(self):
        model = geomodels.data.EModelGroup.ALL

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        n_models = sum(len(m) for m in self.MODEL_TYPES)
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_minimal_models(self):
        model = geomodels.data.EModelGroup.MINIMAL

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        n_models = 3
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_recommended_models(self):
        model = geomodels.data.EModelGroup.RECOMMENDED

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        self.download_mock.assert_called()
        self.assertGreaterEqual(self.download_mock.call_count, 3)
        self.assertLessEqual(self.download_mock.call_count, 7)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertGreaterEqual(self.unpack_archive_mock.call_count, 3)
        self.assertLessEqual(self.unpack_archive_mock.call_count, 7)

    def test_install_geoid_models(self):
        model = geomodels.data.EModelType.GEOID

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        n_models = len(geomodels.data.EGeoidModel)
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertGreaterEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_gravity_models(self):
        model = geomodels.data.EModelType.GRAVITY

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        n_models = len(geomodels.data.EGravityModel)
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertGreaterEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_magnetic_models(self):
        model = geomodels.data.EModelType.MAGNETIC

        with tempfile.TemporaryDirectory() as datadir:
            geomodels.data.install(model, datadir, progress=False)

        n_models = len(geomodels.data.EMagneticModel)
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertGreaterEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_skip(self):
        datadir = geomodels.data.get_default_data_path()
        model = geomodels.data.EMagneticModel.IGRF12

        geomodels.data.install(model, datadir, progress=False)

        self.download_mock.assert_not_called()
        self.unpack_archive_mock.assert_not_called()

    def test_install_skip_one(self):
        model = geomodels.data.EModelType.MAGNETIC

        with tempfile.TemporaryDirectory() as datadir:
            model_dir = pathlib.Path(datadir) / model.value
            model_dir.mkdir()

            igrf12 = geomodels.data.EMagneticModel.IGRF12
            model_file = model_dir / (igrf12.name.lower() + ".dummyext")
            open(model_file, "w").close()  # touch

            geomodels.data.install(model, datadir, progress=False)

        expected_call_count = len(geomodels.data.EMagneticModel) - 1
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, expected_call_count)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertEqual(
            self.unpack_archive_mock.call_count, expected_call_count
        )

    def test_install_invalid_01(self):
        model = None

        with tempfile.TemporaryDirectory() as datadir:
            self.assertRaises(
                (TypeError, ValueError),
                geomodels.data.install,
                model,
                datadir,
                progress=False,
            )



# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

import os
import enum

import numpy as np

cimport cython
from libcpp.string cimport string

from .geoid cimport CGeoid

from .error import GeographicErr
from ._utils import (
    as_contiguous_1d_llh,
    as_contiguous_1d_components,
    reshape_components,
)
from ._typing import PathType

ctypedef CGeoid.convertflag ConvDir


class EHeightConvDir(enum.IntEnum):
    ELLIPSOIDTOGEOID = -1
    NONE = 0
    GEOIDTOELLIPSOID = 1


cdef class GeoidModel:
    """Looking up the height of the geoid above the ellipsoid.

    This class evaluates the height of one of the standard geoids, `EGM84`,
    `EGM96`, or `EGM2008` by bi-linear or cubic interpolation into a
    rectangular grid of data.

    See https://geographiclib.sourceforge.io/html/geoid.html.

    The geoids are defined in terms of spherical harmonics.  However in order
    to provide a quick and flexible method of evaluating the geoid heights,
    this class evaluates the height by interpolation into a grid of
    pre-computed values.

    The height of the geoid above the ellipsoid, `N`, is sometimes called the
    geoid undulation.  It can be used to convert a height above the ellipsoid,
    `h`, to the corresponding height above the geoid (the orthometric height,
    roughly the height above mean sea level), `H`, using the relations::

        h = N + H
        H = -N + h

    This class is typically `not` thread safe in that a single instantiation
    cannot be safely used by multiple threads because of the way the object
    reads the data set and because it maintains a single-cell cache.  If
    multiple threads need to calculate geoid heights they should all construct
    thread-local instantiations.  Alternatively, set the optional
    thread-safe parameter to true in the constructor.  This causes the
    constructor to read all the data into memory and to turn off the
    single-cell caching which results in a Geoid object which is thread
    safe.

    See https://geographiclib.sourceforge.io/html/geoid.html.

    :param name:
        the name of the geoid
    :param path:
        (optional) directory for data file
    :param cubic:
        (optional) interpolation method; false means `bilinear`,
        true (the default) means `cubic`.
    :param threadsafe:
        (optional), if true, construct a thread safe object.
        The default is false
    :raises GeographicErr:
        if the data file cannot be found, is unreadable, or is corrupt
    :raises GeographicErr:
        if `threadsafe` is True but the memory necessary for caching
        the data can't be allocated.

    The data file is formed by appending ".pgm" to the name.
    If `path` is specified (and is non-empty), then the file is loaded
    from directory, `path`.
    Otherwise the path is given by :meth:`default_geoid_path`.
    If the `threadsafe` parameter is True, the data set is read into
    memory, the data file is closed, and single-cell caching is turned
    off; this results in a :class:`Geoid` object which is thread safe.

    Example
    -------
    ::

        # Example of using the `geomodels.GeoidModel` class.
        # This requires that the `egm96-5` geoid model be installed; see
        # https://geographiclib.sourceforge.io/C++/doc/geoid.html#geoidinst

        from geomodels import GeoidModel, EHeightConvDir

        GEOIDTOELLIPSOID = EHeightConvDir.GEOIDTOELLIPSOID

        egm96 = Geoid("egm96-5")

        # Convert height above egm96 to height above the ellipsoid
        lat = 42
        lon = -75
        height_above_geoid = 20

        geoid_height = egm96(lat, lon)
        height_above_ellipsoid = (
            height_above_geoid + GEOIDTOELLIPSOID * geoid_height
        )

        print(f"Height above ellipsoid: {height_above_ellipsoid}")
    """
    cdef CGeoid *_ptr

    def __cinit__(
        self,
        name: str | None = None,
        path: PathType = "",
        *,
        bint cubic=True,
        bint threadsafe=False,
    ):
        cdef string c_name = CGeoid.DefaultGeoidName()
        cdef string c_path = os.fsencode(os.fspath(path))

        if name is not None:
            c_name = os.fsencode(name)

        try:
            with nogil:
                self._ptr = new CGeoid(c_name, c_path, cubic, threadsafe)
        except RuntimeError as exc:
            raise GeographicErr(str(exc)) from exc

    def __dealloc__(self):
        del self._ptr

    def cache_area(self, south: float, west: float, north: float, east: float):
        """Set up a cache.

        :param float south:
            south latitude (degrees) of the south edge of the cached area.
        :param west:
            west longitude (degrees) of the west edge of the cached area.
        :param north:
            north latitude (degrees) of the north edge of the cached area.
        :param east:
            east longitude (degrees) of the east edge of the cached area.
        :raises GeographicErr:
            if the memory necessary for caching the data can't be
            allocated (in this case, you will have no cache and can try
            again with a smaller area).
        :raises GeographicErr:
            if there's a problem reading the data.
        :raises GeographicErr:
            if this is called on a thread-safe :class:`Geoid`.

        Cache the data for the specified "rectangular" area bounded by
        the parallels `south` and `north` and the meridians `west` and
        `east`.
        `east` is always interpreted as being east of `west`, if
        necessary by adding 360 deg; to its value.
        `south` and `north` should be in the range [-90 deg, +90 deg].
        """
        try:
            with nogil:
                self._ptr.CacheArea(south, west, north, east)
        except RecursionError as exc:
            raise GeographicErr(str(exc)) from exc

    def cache_all(self):
        """Cache all the data.

        :raises GeographicErr:
            if the memory necessary for caching the data can't be
            allocated (in this case, you will have no cache and can try
            again with a smaller area).
        :raises GeographicErr:
            if there's a problem reading the data.
        :raises GeographicErr:
            if this is called on a thread-safe :class:`Geoid`.

        On most computers, this is fast for data sets with grid
        resolution of 5' or coarser. For a 1' grid, the required RAM is
        450MB; a 2.5' grid needs 72MB; and a 5' grid needs 18MB.
        """
        try:
            with nogil:
                self._ptr.CacheAll()
        except RecursionError as exc:
            raise GeographicErr(str(exc)) from exc

    def cache_clear(self):
        """Clear the cache.

        This never throws an error.
        This does nothing with a thread safe :class:`Geoid`.
        """
        self._ptr.CacheClear()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _compute(self, double[::1] vlat, double[::1] vlon):
        cdef long size = vlat.size
        h = np.empty(shape=[size], dtype=np.float64)
        cdef double[::1] vh = h

        try:
            with nogil:
                for i in range(size):
                    vh[i] = cython.operator.dereference(self._ptr)(
                        vlat[i], vlon[i])
        except RecursionError as exc:
            raise GeographicErr(str(exc)) from exc

        return h

    def __call__(self, lat, lon):
        """Compute the geoid height at a point.

        :param lat:
            latitude of the point (degrees).
        :param lon:
            longitude of the point (degrees).
        :raises GeographicErr:
            if there's a problem reading the data.
            This never happens if (`lat`, `lon`) is within a
            successfully cached area.
        :returns:
            the height of the geoid above the ellipsoid (meters).

        The latitude should be in [-90 deg, +90deg].
        """
        dtype = np.float64
        lat, lon, shape = as_contiguous_1d_components(
            lat, lon, labels=['lat', 'lon'], dtype=dtype
        )
        h = self._compute(lat, lon)
        return reshape_components(shape, h)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _convert_height(
        self,
        double[::1] vlat,
        double[::1] vlon,
        double[::1] vh,
        ConvDir direction,
    ):
        cdef long size = vlat.size
        out = np.empty(shape=[size], dtype=np.float64)
        cdef double[::1] vout = out
        cdef long i = 0

        try:
            with nogil:
                for i in range(size):
                    vout[i] = self._ptr.ConvertHeight(
                        vlat[i], vlon[i], vh[i], direction
                    )
        except RecursionError as exc:
            raise GeographicErr(str(exc)) from exc

        return out

    def convert_height(self, lat, lon, h, direction: EHeightConvDir):
        """Convert a height above the geoid to a height above the
        ellipsoid and vice versa.

        :param lat:
            latitude of the point (degrees).
        :param lon:
            longitude of the point (degrees).
        :param h:
            height of the point (meters).
        :param direction:
            a :class:`EHeightConvDir` specifying the direction of the
            conversion; :data:`EHeightConvDir.GEOIDTOELLIPSOID` means
            convert a height above the geoid to a height above the
            ellipsoid; :data:`Geoid.ELLIPSOIDTOGEOID` means convert a
            height above the ellipsoid to a height above the geoid.
        :raises GeographicErr:
            if there's a problem reading the data; this never happens
            if (`lat`, `lon`) is within a successfully cached area.
        :returns:
            converted height (meters).
        """
        direction = EHeightConvDir(direction)
        dtype = np.float64
        lat, lon, h, shape = as_contiguous_1d_llh(lat, lon, h, dtype)

        out = self._convert_height(lat, lon, h, direction.value)

        return reshape_components(shape, out)

    def description(self) -> str:
        """Return geoid description, if available, in the data file.

        If the geoid description is absent in the data file
        return "NONE".
        """
        return self._ptr.Description().decode('utf-8')

    def datetime(self) -> str:
        """Return date of the data file.

        If absent, return "UNKNOWN".
        """
        return self._ptr.DateTime().decode('utf-8')

    def geoid_file(self) -> str:
        """Return full file name used to load the geoid data."""
        return self._ptr.GeoidFile().decode('utf-8')

    def geoid_name(self) -> str:
        """Return the "name" used to load the geoid data.

        "name" is the first argument of the constructor.
        """
        return self._ptr.GeoidName().decode('utf-8')

    def geoid_directory(self) -> str:
        """Return the directory used to load the geoid data."""
        return self._ptr.GeoidDirectory().decode('utf-8')

    def interpolation(self) -> str:
        """Return interpolation method (`cubic` or `bilinear`)."""
        # @TODO: use an enum
        return self._ptr.Interpolation().decode('utf-8')

    def max_error(self) -> float:
        """Return an estimate of the maximum interpolation and
        quantization error [m].

        This relies on the value being stored in the data file.
        If the value is absent, return -1.
        """
        return self._ptr.MaxError()

    def rms_error(self) -> float:
        """Return an estimate of the RMS interpolation and
        quantization error [m].

        This relies on the value being stored in the data file.
        If the value is absent, return -1.
        """
        return self._ptr.RMSError()

    def offset(self) -> float:
        """Offset (meters).

        This in used in converting from the pixel values in the data
        file to geoid heights.
        """
        return self._ptr.Offset()

    def scale(self) -> float:
        """Scale (meters).

        This in used in converting from the pixel values in the data
        file to geoid heights.
        """
        return self._ptr.Scale()

    def threadsafe(self) -> bool:
        """Return true if the object is constructed to be thread safe."""
        return self._ptr.ThreadSafe()

    def cache(self) -> bool:
        """Return true if a data cache is active."""
        return self._ptr.Cache()

    def cache_west(self) -> float:
        """West edge of the cached area.

        The cache includes this edge.
        """
        return self._ptr.CacheWest()

    def cache_east(self) -> float:
        """East edge of the cached area.

        The cache excludes this edge.
        """
        return self._ptr.CacheEast()

    def cache_north(self) -> float:
        """North edge of the cached area.

        The cache includes this edge.
        """
        return self._ptr.CacheNorth()

    def cache_south(self) -> float:
        """South edge of the cached area.

        The cache excludes this edge unless it's the south pole.
        """
        return self._ptr.CacheSouth()

    def equatorial_radius(self) -> float:
        """The equatorial radius of the ellipsoid (meters).

        (The WGS84 value is returned because the supported geoid
        models are all based on this ellipsoid.)
        """
        return self._ptr.EquatorialRadius()

    def flattening(self) -> float:
        """The flattening of the ellipsoid.

        (The WGS84 value is returned because the supported geoid
        models are all based on this ellipsoid.
        """
        return self._ptr.Flattening()

    @staticmethod
    def default_geoid_path() -> str:
        """The default path for geoid data files.

        This is the value of the environment variable
        `GEOGRAPHICLIB_GEOID_PATH`, if set; otherwise, it is
        `${GEOGRAPHICLIB_DATA}/geoids` if the environment variable
        `GEOGRAPHICLIB_DATA` is set; otherwise, it is a compile-time
        default.
        """
        return CGeoid.DefaultGeoidPath().decode('utf-8')

    @staticmethod
    def default_geoid_name() -> str:
        """The default name for the geoid.

        This is the value of the environment variable
        `GEOGRAPHICLIB_GEOID_NAME`, if set; otherwise, it is "egm96-5".
        The GeoidModel class does not use this function; it is just
        provided as a convenience for a calling program when
        constructing a :class:`GeoidModel` object.
        """
        return CGeoid.DefaultGeoidName().decode('utf-8')



import re
import unittest

import geomodels


class GeographicLibCommonCase(unittest.TestCase):
    def test_lib_version_str(self):
        version = geomodels.lib_version_str()
        self.assertIsInstance(version, str)
        self.assertRegex(version, r"\d+\.\d+(\.\d+)?")

    def test_lib_version_info(self):
        info = geomodels.lib_version_info()
        self.assertIsInstance(info, tuple)
        self.assertEqual(len(info), 3)
        self.assertEqual(info, (info.major, info.minor, info.micro))
        self.assertIsInstance(info.major, int)
        self.assertIsInstance(info.minor, int)
        self.assertIsInstance(info.micro, int)

    def test_version_consistency(self):
        version_str = geomodels.lib_version_str()
        version_info = geomodels.lib_version_info()

        mobj = re.match(
            r"(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<micro>\d+))?", version_str
        )
        if mobj:
            self.assertEqual(version_info.major, int(mobj.group("major")))
            self.assertEqual(version_info.minor, int(mobj.group("minor")))
            if mobj.group("micro"):
                self.assertEqual(version_info.micro, int(mobj.group("micro")))
        else:
            raise RuntimeError(f"invalid version string: {version_str!r}")



import enum
from typing import NamedTuple

import numpy as np
from numpy import typing as npt

from ._typing import PathType

class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int

def lib_version_info() -> VersionInfo: ...
def lib_version_str() -> str: ...

class EHeightConvDir(enum.IntEnum):
    ELLIPSOIDTOGEOID = -1
    NONE = 0
    GEOIDTOELLIPSOID = 1

class GeoidModel:
    def __init__(
        self,
        name: str | None = None,
        path: PathType = "",
        *,
        cubic: bool = True,
        threadsafe: bool = False,
    ): ...
    def cache_area(
        self, south: float, west: float, north: float, east: float
    ) -> None: ...
    def cache_all(self) -> None: ...
    def cache_clear(self) -> None: ...
    def __call__(
        self, lat: npt.ArrayLike, lon: npt.ArrayLike
    ) -> np.ndarray: ...
    def convert_height(
        self,
        lat: npt.ArrayLike,
        lon: npt.ArrayLike,
        h: npt.ArrayLike,
        direction: EHeightConvDir,
    ) -> np.ndarray: ...
    def description(self) -> str: ...
    def datetime(self) -> str: ...
    def geoid_file(self) -> str: ...
    def geoid_name(self) -> str: ...
    def geoid_directory(self) -> str: ...
    def interpolation(self) -> str: ...
    def max_error(self) -> float: ...
    def rms_error(self) -> float: ...
    def offset(self) -> float: ...
    def scale(self) -> float: ...
    def threadsafe(self) -> bool: ...
    def cache(self) -> bool: ...
    def cache_west(self) -> float: ...
    def cache_east(self) -> float: ...
    def cache_north(self) -> float: ...
    def cache_south(self) -> float: ...
    def equatorial_radius(self) -> float: ...
    def flattening(self) -> float: ...
    @staticmethod
    def default_geoid_path() -> str: ...
    @staticmethod
    def default_geoid_name() -> str: ...

class GravityModel:
    def __init__(
        self,
        name: str | None = None,
        path: PathType = "",
        max_degree: int = -1,
        max_order: int = -1,
    ): ...
    def gravity(
        self, lat: npt.ArrayLike, lon: npt.ArrayLike, h: npt.ArrayLike
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def disturbance(
        self,
        lat: npt.ArrayLike,
        lon: npt.ArrayLike,
        h: npt.ArrayLike,
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def geoid_height(
        self, lat: npt.ArrayLike, lon: npt.ArrayLike
    ) -> np.ndarray: ...
    def spherical_anomaly(
        self, lat: npt.ArrayLike, lon: npt.ArrayLike, h: npt.ArrayLike
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def w(
        self, lat: npt.ArrayLike, lon: npt.ArrayLike, h: npt.ArrayLike
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def v(self, x: npt.ArrayLike, y: npt.ArrayLike, z: npt.ArrayLike) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def t_components(
        self, x: npt.ArrayLike, y: npt.ArrayLike, z: npt.ArrayLike
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def t(
        self, x: npt.ArrayLike, y: npt.ArrayLike, z: npt.ArrayLike
    ) -> np.ndarray: ...
    def u(self, x: npt.ArrayLike, y: npt.ArrayLike, z: npt.ArrayLike) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def phi(self, x: npt.ArrayLike, y: npt.ArrayLike) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def description(self) -> str: ...
    def datetime(self) -> str: ...
    def gravity_file(self) -> str: ...
    def gravity_model_name(self) -> str: ...
    def gravity_model_directory(self) -> str: ...
    def equatorial_radius(self) -> float: ...
    def flattening(self) -> float: ...
    def mass_constant(self) -> float: ...
    def reference_mass_constant(self) -> float: ...
    def angular_velocity(self) -> float: ...
    def degree(self) -> int: ...
    def order(self) -> int: ...
    @staticmethod
    def default_gravity_path() -> str: ...
    @staticmethod
    def default_gravity_name() -> str: ...

class MagneticFieldModel:
    def __init__(self, name: str | None = None, path: PathType = ""): ...
    def __call__(
        self,
        t: float,
        lat: npt.ArrayLike,
        lon: npt.ArrayLike,
        h: npt.ArrayLike,
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def compute_with_rate(
        self,
        t: float,
        lat: npt.ArrayLike,
        lon: npt.ArrayLike,
        h: npt.ArrayLike,
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    @staticmethod
    def field_components(
        Bx: npt.ArrayLike, By: npt.ArrayLike, Bz: npt.ArrayLike
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]: ...
    def description(self) -> str: ...
    def datetime(self) -> str: ...
    def magnetic_file(self) -> str: ...
    def magnetic_model_name(self) -> str: ...
    def magnetic_model_directory(self) -> str: ...
    def min_height(self) -> float: ...
    def max_height(self) -> float: ...
    def min_time(self) -> float: ...
    def max_time(self) -> float: ...
    def equatorial_radius(self) -> float: ...
    def flattening(self) -> float: ...
    def degree(self) -> int: ...
    def order(self) -> int: ...
    @staticmethod
    def default_magnetic_path() -> str: ...
    @staticmethod
    def default_magnetic_name() -> str: ...



#!/usr/bin/env python3

import os
import re
import sys
import pathlib
import platform

from setuptools import setup, Extension


def mkconfig(srcpath, outpath, have_lohg_double=0):
    data = pathlib.Path(srcpath).joinpath("CMakeLists.txt").read_text()
    mobj = re.search(
        r"""\
set\s*\(\s*PROJECT_VERSION_MAJOR\s+(?P<VERSION_MAJOR>\d+)\s*\)\s*
set\s*\(\s*PROJECT_VERSION_MINOR\s+(?P<VERSION_MINOR>\d+)\s*\)\s*
set\s*\(\s*PROJECT_VERSION_PATCH\s+(?P<VERSION_PATCH>\d+)\s*\)\s*
""",
        data,
        re.MULTILINE,
    )

    version_major = mobj.group("VERSION_MAJOR")
    version_minor = mobj.group("VERSION_MINOR")
    version_patch = mobj.group("VERSION_PATCH")

    if version_patch == "0":
        version_string = ".".join([version_major, version_minor])
    else:
        version_string = ".".join(
            [version_major, version_minor, version_patch],
        )

    bigendian = 0 if sys.byteorder == "little" else 1

    configdata = f"""\
#define GEOGRAPHICLIB_VERSION_STRING "{version_string}"
#define GEOGRAPHICLIB_VERSION_MAJOR {version_major}
#define GEOGRAPHICLIB_VERSION_MINOR {version_minor}
#define GEOGRAPHICLIB_VERSION_PATCH {version_patch}
#define GEOGRAPHICLIB_DATA "/usr/local/share/GeographicLib"

// These are macros which affect the building of the library
#define GEOGRAPHICLIB_HAVE_LONG_DOUBLE {have_lohg_double}
#define GEOGRAPHICLIB_WORDS_BIGENDIAN {bigendian}
#define GEOGRAPHICLIB_PRECISION 2

#if !defined(GEOGRAPHICLIB_SHARED_LIB)
#define GEOGRAPHICLIB_SHARED_LIB 0
#endif
"""

    outpath = pathlib.Path(outpath)
    if not outpath.exists():
        outpath.write_text(configdata)
        print(f"configuration file written to: {outpath}")
    else:
        print(f"configuration file already exists: {outpath}")


if os.name == "posix":
    if platform.system() == "FreeBSD":
        mandir = pathlib.Path("man")
    else:
        mandir = pathlib.Path("share/man")
    datafiles = [
        (
            str(mandir / "man1"),
            ["docs/man/geomodels-cli.1"],
        )
    ]
else:
    datafiles = []


IGNORE_BUNDLED_LIBS_STR = os.environ.get("GEOMODELS_IGNORE_BUNDLED_LIBS")
IGNORE_BUNDLED_LIBS = bool(
    IGNORE_BUNDLED_LIBS_STR in ("1", "ON", "TRUE", "YES")
)
SRCPATH = pathlib.Path("extern/geographiclib")


if SRCPATH.exists() and not IGNORE_BUNDLED_LIBS:
    geographiclib_src = SRCPATH.glob("src/*.cpp")
    geographiclib_include = list(SRCPATH.glob("include"))[0]
    geomodels_ext = Extension(
        "geomodels._ext",
        sources=["geomodels/_ext.pyx"] + [str(p) for p in geographiclib_src],
        include_dirs=[str(geographiclib_include)],
        libraries=[],
        language="c++",
        extra_compile_args=["-std=c++11", "-Wall"],
    )
    outpath = geographiclib_include / "GeographicLib" / "Config.h"
    mkconfig(SRCPATH, outpath)
else:
    libname = os.environ.get("GEOMODELS_GEOGRAPHICLIB_NAME", "GeographicLib")
    geomodels_ext = Extension(
        "geomodels._ext",
        sources=["geomodels/_ext.pyx"],
        libraries=[libname],
        language="c++",
    )


description = pathlib.Path("README.rst").read_text()
description = description.replace(".. doctest", "").replace(":doc:", "")


setup(
    long_description=description,
    long_description_content_type="text/x-rst",
    ext_modules=[geomodels_ext],
    data_files=datafiles,
)



"""Unit tests for the geomodels package."""

import os
import sys
import locale
import platform
import unittest


def print_versions():
    """Print platform information and library version."""
    from .. import lib_version_str, __version__  # avoid circular imports

    geographiclib_data = os.environ.get("GEOGRAPHICLIB_DATA", "not specified")

    print(f"geomodels version:     {__version__}")
    print(f"GeographicLib version: {lib_version_str()}")
    print(f"GEOGRAPHICLIB_DATA:    {geographiclib_data}")

    print(f"Python version:        {platform.python_version()}")
    print(f"Platform:              {platform.platform()}")
    print(f"Byte-ordering:         {sys.byteorder}")
    print(f"Default encoding:      {sys.getdefaultencoding()}")
    print(f"Default FS encoding:   {sys.getfilesystemencoding()}")
    print(f"Locale:                {locale.getlocale()}")

    print()

    sys.stdout.flush()


def suite():
    """Return the test suite for the geomodels package."""
    loader = unittest.TestLoader()
    return loader.discover(start_dir=os.path.dirname(__file__))


def test(verbosity: int = 1, failfast: bool = False):
    """Run the test suite for the geomodels package.

    :param int verbosity:
        verbosity level (higher is more verbose).
        Default: 1.
    :param bool failfast:
        stop the test run on the first error or failure.
        Default: False.
    """
    print_versions()
    runner = unittest.TextTestRunner(verbosity=verbosity, failfast=failfast)
    result = runner.run(suite())

    return os.EX_OK if result.wasSuccessful() else 1



import ctypes

intersect_dso = ctypes.CDLL('./intersect_dso.so')
intersect_c = intersect_dso.intersect
intersect_c.argtypes = ([ctypes.c_double] * 6) + ([ctypes.POINTER(ctypes.c_double)] * 2)

def intersect(lat1, lon1, az1, lat2, lon2, az2):
    dX = ctypes.c_double()
    dY = ctypes.c_double()

    intersect_c(
        lat1, lon1, az1, lat2, lon2, az2,
        ctypes.byref(dX), ctypes.byref(dY),
    )

    return dX.value, dY.value



import os
import copy
import shutil
import pathlib
import tempfile
import unittest
from typing import Protocol

import numpy as np
from numpy import testing as npt

import geomodels
from geomodels import wmmf

DATAPATH = pathlib.Path(geomodels.MagneticFieldModel.default_magnetic_path())


class MetaDataTestProtocol(Protocol):
    @property
    def metadata(self) -> wmmf.MetaData: ...

    def assertEqual(self, *args, **kargs): ...  # noqa: N802
    def assertIsInstance(self, *args, **kargs): ...  # noqa: N802
    def assertTrue(self, *args, **kargs): ...  # noqa: N802
    def subTest(self, *args, **kargs): ...  # noqa: N802


class MetaDataTestMixin:
    def test_fields(self: MetaDataTestProtocol):
        for name in (
            "Name",
            "Description",
            "URL",
            "Publisher",
            "ReleaseDate",
            "DataCutOff",
            "ConversionDate",
            "ID",
        ):
            with self.subTest(name=name):
                self.assertTrue(hasattr(self.metadata, name))
                self.assertIsInstance(getattr(self.metadata, name), str)

        for name in (
            "FORMAT_VERSION",
            "NumModels",
            "Epoch",
            "DeltaEpoch",
            "MinTime",
            "MaxTime",
            "DataVersion",
            "Radius",
            "MinHeight",
            "MaxHeight",
        ):
            with self.subTest(name=name):
                self.assertTrue(hasattr(self.metadata, name))
                self.assertIsInstance(getattr(self.metadata, name), int)

    def test_get_id(self: MetaDataTestProtocol):
        self.assertEqual(self.metadata.get_id(), self.metadata.ID)


class EmptyMetaDataTestCase(MetaDataTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.metadata = wmmf.MetaData()

    def test_get_years(self):
        self.assertEqual(self.metadata.get_years(), [])


class MetaDataTestCase(MetaDataTestMixin, unittest.TestCase):
    METADATAFILE = DATAPATH / "igrf12.wmm"

    def setUp(self) -> None:
        self.metadata = wmmf.MetaData()
        self.metadata.load(self.METADATAFILE)

    def test_load(self):
        self.assertEqual(self.metadata.FORMAT_VERSION, 1)
        self.assertEqual(self.metadata.Name, "igrf12")
        self.assertEqual(self.metadata.NumModels, 24)
        self.assertEqual(self.metadata.Epoch, 1900)
        self.assertEqual(self.metadata.DeltaEpoch, 5)
        self.assertEqual(self.metadata.MinTime, 1900)
        self.assertEqual(self.metadata.MaxTime, 2020)

    def test_get_years(self):
        years = self.metadata.get_years()
        self.assertEqual(len(years), self.metadata.NumModels)
        for year in years:
            with self.subTest(year=year):
                self.assertIsInstance(year, int)

    # TODO: make this test more robust
    def test_str(self):
        with open(self.METADATAFILE) as fd:
            data = fd.read().strip()
        self.assertEqual(data, str(self.metadata).strip())

    def test_save(self):
        with tempfile.TemporaryDirectory() as outpath:
            filename = pathlib.Path(outpath) / "igrf12.wmm"
            self.metadata.save(filename)
            self.assertTrue(os.path.exists(filename))

            local_metadata = wmmf.MetaData()
            local_metadata.load(filename)
            self.assertEqual(self.metadata, local_metadata)


class MetaDataFromTxtHeaderTestCase(MetaDataTestMixin, unittest.TestCase):
    COEFFFILE = pathlib.Path(__file__).parent / "data" / "igrf12coeffs.txt"

    def setUp(self) -> None:
        with open(self.COEFFFILE) as fd:
            data = fd.read()
        header = data.splitlines(keepends=False)[3]
        self.metadata = wmmf._metadata_from_txt_header(header)
        self.metadata.Name = "igrf12"

    def test_from_header(self):
        self.assertEqual(self.metadata.FORMAT_VERSION, 1)
        self.assertEqual(self.metadata.Name, "igrf12")
        self.assertEqual(self.metadata.NumModels, 24)
        self.assertEqual(self.metadata.Epoch, 1900)
        self.assertEqual(self.metadata.DeltaEpoch, 5)
        self.assertEqual(self.metadata.MinTime, 1900)
        self.assertEqual(self.metadata.MaxTime, 2020)

    def test_get_years(self):
        years = self.metadata.get_years()
        self.assertEqual(len(years), self.metadata.NumModels)
        for year in years:
            with self.subTest(year=year):
                self.assertIsInstance(year, int)

    def test_get_id(self):
        self.assertEqual(self.metadata.get_id(), "IGRF12-A")


class WmmDataTestCase(unittest.TestCase):
    METADATAFILE = DATAPATH / "igrf12.wmm"

    def setUp(self) -> None:
        self.wmmdata = wmmf.WmmData(self.METADATAFILE)

    def test_load_metadata(self):
        metadata = wmmf.MetaData()
        metadata.load(self.METADATAFILE)
        self.assertEqual(self.wmmdata.metadata, metadata)

    def test_load_coeffs(self):
        self.assertNotEqual(self.wmmdata.coeffs, {})
        self.assertIn("rate", self.wmmdata.coeffs)
        self.assertEqual("rate", list(self.wmmdata.coeffs.keys())[-1])

    def test_load_consistency(self):
        self.assertEqual(
            len(self.wmmdata.coeffs), self.wmmdata.metadata.NumModels + 1
        )
        years = [
            int(item) for item in self.wmmdata.coeffs.keys() if item != "rate"
        ]
        self.assertEqual(years, self.wmmdata.metadata.get_years())

    def test_coeffs_data(self):
        coeffs = self.wmmdata.coeffs
        for year, sph_coeff_set in coeffs.items():
            with self.subTest(year=year):
                self.assertIsInstance(year, str)
                self.assertEqual(sph_coeff_set.C.dtype, np.dtype("<f8"))
                self.assertEqual(sph_coeff_set.S.dtype, np.dtype("<f8"))
                self.assertEqual(sph_coeff_set.C.shape, sph_coeff_set.S.shape)

    def test_save_to_dir(self):
        with tempfile.TemporaryDirectory() as outpath:
            metadata_filename = pathlib.Path(outpath) / "igrf12.wmm"
            bin_filename = metadata_filename.with_suffix(".wmm.cof")

            self.assertFalse(metadata_filename.is_file())
            self.assertFalse(bin_filename.is_file())

            self.wmmdata.save(outpath)

            self.assertTrue(metadata_filename.is_file())
            self.assertTrue(bin_filename.is_file())

    def test_save_to_file(self):
        with tempfile.TemporaryDirectory() as outpath:
            metadata_filename = pathlib.Path(outpath) / "custom_igrf12.wmm"
            bin_filename = metadata_filename.with_suffix(".wmm.cof")

            self.assertFalse(metadata_filename.is_file())
            self.assertFalse(bin_filename.is_file())

            self.wmmdata.save(metadata_filename)

            self.assertTrue(metadata_filename.is_file())
            self.assertTrue(bin_filename.is_file())

    def test_from_metadata_and_coeffs(self):
        src_wmmdata = self.wmmdata
        dst_wmmdata = wmmf.WmmData.from_metadata_and_coeffs(
            src_wmmdata.metadata, src_wmmdata.coeffs
        )
        self.assertEqual(src_wmmdata.metadata, dst_wmmdata.metadata)
        self.assertEqual(src_wmmdata.coeffs, dst_wmmdata.coeffs)

    def test_from_inconsistent_metadata_and_coeffs_01(self):
        src_wmmdata = self.wmmdata
        metadata = copy.copy(src_wmmdata.metadata)
        metadata.NumModels = metadata.NumModels + 2
        self.assertRaises(
            RuntimeError,
            wmmf.WmmData.from_metadata_and_coeffs,
            metadata,
            src_wmmdata.coeffs,
        )

    def test_from_inconsistent_metadata_and_coeffs_02(self):
        src_wmmdata = self.wmmdata
        metadata = copy.copy(src_wmmdata.metadata)
        metadata.NumModels = metadata.NumModels - 2
        self.assertRaises(
            RuntimeError,
            wmmf.WmmData.from_metadata_and_coeffs,
            metadata,
            src_wmmdata.coeffs,
        )

    def test_from_inconsistent_metadata_and_coeffs_03(self):
        coeffs = self.wmmdata.coeffs.copy()
        coeffs.pop("rate")
        self.assertRaises(
            RuntimeError,
            wmmf.WmmData.from_metadata_and_coeffs,
            self.wmmdata.metadata,
            coeffs,
        )

    def test_from_inconsistent_metadata_and_coeffs_04(self):
        coeffs = self.wmmdata.coeffs.copy()
        year = "1950"
        item = coeffs.pop(year)
        coeffs[year] = item
        item = coeffs.pop("rate")
        coeffs["tare"] = item
        self.assertEqual(list(coeffs.keys())[-2], year)
        self.assertRaises(
            RuntimeError,
            wmmf.WmmData.from_metadata_and_coeffs,
            self.wmmdata.metadata,
            coeffs,
        )


class ImportIgrfTxtTestCase(WmmDataTestCase):
    IGRFTXT = pathlib.Path(__file__).parent / "data" / "igrf12coeffs.txt"

    def setUp(self) -> None:
        self.wmmdata = wmmf.import_igrf_txt(self.IGRFTXT)

    def test_load_metadata(self):
        metadata = wmmf.MetaData()
        metadata.load(self.METADATAFILE)
        for name in vars(self.wmmdata.metadata):
            with self.subTest(name=name):
                if name.startswith("_"):
                    continue
                if name in ("Description", "ConversionDate"):
                    continue
                value = getattr(self.wmmdata.metadata, name)
                if value == "N/A":
                    continue
                self.assertEqual(value, getattr(metadata, name))


NO_NETWORK = bool(
    os.environ.get("GEOMODELS_NO_NETWORK") in ("1", "OK", "TRUE", "YES")
)


@unittest.skipIf(NO_NETWORK, "no network")
class ImportRemoteIgrfTxtTestCase(WmmDataTestCase):
    IGRFTXT = "https://www.ngdc.noaa.gov/IAGA/vmod/coeffs/igrf13coeffs.txt"


class DataCrossCheckTestCase(unittest.TestCase):
    METADATAFILE = DATAPATH / "igrf12.wmm"
    IGRFTXT = pathlib.Path(__file__).parent / "data" / "igrf12coeffs.txt"

    def setUp(self) -> None:
        self.tmpdir = tempfile.mkdtemp()

        wmmtxt = wmmf.import_igrf_txt(self.IGRFTXT)
        wmmtxt.save(self.tmpdir)

        filename = pathlib.Path(self.tmpdir) / "igrf12.wmm"
        self.wmmtxt = wmmf.WmmData(filename)
        self.wmmbin = wmmf.WmmData(self.METADATAFILE)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir)

    def test_coeffs_keys(self):
        self.assertEqual(
            list(self.wmmbin.coeffs.keys()), list(self.wmmtxt.coeffs.keys())
        )

    def test_data(self):
        it = zip(
            self.wmmbin.coeffs.keys(),
            self.wmmbin.coeffs.values(),
            self.wmmtxt.coeffs.values(),
        )
        for year, bin_, txt in it:
            with self.subTest(year=year):
                self.assertEqual(bin_.C.shape, txt.C.shape)
                self.assertEqual(bin_.S.shape, txt.S.shape)
                npt.assert_allclose(bin_.C, txt.C)
                npt.assert_allclose(bin_.S, txt.S)

    def test_bin_filesize(self):
        orig = pathlib.Path(self.METADATAFILE).with_suffix(".wmm.cof")
        new = pathlib.Path(self.tmpdir) / "igrf12.wmm.cof"
        self.assertEqual(orig.stat().st_size, new.stat().st_size)

    def test_binary_data(self):
        orig = pathlib.Path(self.METADATAFILE).with_suffix(".wmm.cof")
        new = pathlib.Path(self.tmpdir) / "igrf12.wmm.cof"
        self.assertEqual(orig.read_bytes(), new.read_bytes())



# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

from libcpp cimport bool
from libcpp.string cimport string

from .common cimport real


cdef extern from "GeographicLib/Geoid.hpp" namespace "GeographicLib" nogil:
    cdef cppclass CGeoid "GeographicLib::Geoid":
        ctypedef enum convertflag:
            ELLIPSOIDTOGEOID = -1,
            NONE = 0,
            GEOIDTOELLIPSOID = 1

        CGeoid(
            const string& name,
            const string& path,
            bool cubic,
            bool threadsafe,
        ) except +

        void CacheArea(real south, real west, real north, real east) except +  # const
        void CacheAll() except +  # const
        void CacheClear() const

        real operator()(real lat, real lon) except +  # const
        real ConvertHeight(real lat, real lon, real h, convertflag d) except +  # const

        const string& Description() const
        const string& DateTime() const
        const string& GeoidFile() const
        const string& GeoidName() const
        const string& GeoidDirectory() const
        const string Interpolation() const

        real MaxError() const
        real RMSError() const
        real Offset() const
        real Scale() const

        bool ThreadSafe() const

        bool Cache() const
        real CacheWest() const
        real CacheEast() const
        real CacheNorth() const
        real CacheSouth() const

        real EquatorialRadius() const
        real Flattening() const

        @staticmethod
        string DefaultGeoidPath()

        @staticmethod
        string DefaultGeoidName()
